#include<stdio.h>
#include<time.h>
#include<stdlib.h>

typedef unsigned int uint;

const int LMASK=8; // 256 bits
const int LBASE=1000000;

uint Base[LBASE][LMASK];
uint Tree[2*LBASE][2];   // equal bits:offset in base:ref to "1"
char NBit[256];

extern "C" int _bcmp(const void *a,const void *b){
	const uint *A=(const uint *)a;	
	const uint *B=(const uint *)b;
	for(int i=LMASK;--i>=0;){
		if(A[i]<B[i]) return -1;		
		if(A[i]>B[i]) return 1;
	}
	return 0;
}



uint seed=1234567;
int irand(int n){
	seed=seed*0x1010005+1;
	long long m=(long long)seed*n;
	return (int)(m>>32);
}

void xorbit(uint *m,int k){
	m[k>>5]^=(1u<<(k&31));
}


int BuildTree(int ptree,int nbit,int bstart,int bend){
	if(nbit!=0 && bend!=bstart+1){
		for(int nb=nbit-1;nb>=0;nb--){
			int pb=nb>>5,mb=1<<(nb&31);
			uint u=Base[bstart][pb];
			int qdif=-1;
			for(int i=bstart+1;i<bend;i++){
				if((Base[i][pb]^u)&mb){ qdif=i; break; }
			}
			if(qdif>=0){
				Tree[ptree][0]=bstart|((nbit-nb-1)<<24);
				int h=BuildTree(ptree+1,nb,bstart,qdif);
				Tree[ptree][1]=h;
				return BuildTree(h,nb,qdif,bend);
			}
		}
	}
	Tree[ptree][0]=bstart+(nbit<<24);
	return ptree+1;
}

void PrepareTree(int lbase){
	qsort(Base,lbase,LMASK*4,_bcmp);
	int ptree=0;
	int h=BuildTree(ptree,LMASK*32,0,lbase);
	NBit[0]=0;
	for(int m=1;m<256;m*=2) for(int i=0;i<m;i++) NBit[i+m]=NBit[i]+1;
}

int Res[LBASE],nres;


int HamDist(uint *a,uint *b,int b0,int b1){
	unsigned char *ua=(unsigned char *)a,*ub=(unsigned char *)b;

	int p0=b0>>3,p1=b1>>3;
	ua+=p0; ub+=p0;
	b0&=7; b1&=7;
	if(p0==p1) return NBit[(*ua^*ub)&((1<<b1)-1)&((-1)<<b0)];

	int res=NBit[(*ua++^*ub++)&((-1)<<b0)];
	while(++p0<p1) res+=NBit[*ua++^*ub++];
	return res+NBit[(*ua^*ub)&((1<<b1)-1)];
}

void Find1(uint *targ,int ptree,int nbit,int dist){
	for(;;){
		uint a=Tree[ptree][0];
		int nb=(int)(a>>24); a&=0xffffff;
		if(nb!=0){
			int lh=HamDist(Base[a],targ,nbit-nb,nbit);
			if(lh>dist) return;
			dist-=lh; nbit-=nb;
		}
		if(nbit==0){ Res[nres++]=a; return; }
		nbit--;
		uint u1=targ[nbit>>5]&(1u<<(nbit&31));
		if(dist!=0){
			Find1(targ,u1 ? ptree+1 : Tree[ptree][1],nbit,dist-1);
		}
		ptree=u1 ? Tree[ptree][1]: ptree+1;
	}
}
void Find(uint *targ,int dist){
	nres=0;
	Find1(targ,0,LMASK*32,dist);
}



void FillBase(int l,int dnb){
	for(int i=0;i<LMASK*16;i++) xorbit(Base[0],irand(LMASK*32));
	for(int k=1;k<l;k++){
		int c=irand(k);
		for(int i=0;i<LMASK;i++) Base[k][i]=Base[c][i];
		for(int u=0;u<dnb;u++) xorbit(Base[k],irand(LMASK*32));
	}
}


void main(){
	FillBase(LBASE,1);
	PrepareTree(LBASE);

	int N=0;
	printf("Start: %ld\n",clock());
	for(int i=0;i<1000;i++){
		Find(Base[i*100],15);
		N+=nres;
	}
	printf("End: %ld, nres=%d\n",clock(),N);
}